@@ -14,6 +14,7 @@ import re
 import socket
 import struct
 import sys
+import traceback
 import warnings
 from collections import defaultdict
 from collections import namedtuple
@@ -23,12 +24,16 @@ from . import _psposix
 from . import _psutil_linux as cext
 from . import _psutil_posix as cext_posix
 from ._common import isfile_strict
+from ._common import memoize
+from ._common import parse_environ_block
 from ._common import NIC_DUPLEX_FULL
 from ._common import NIC_DUPLEX_HALF
 from ._common import NIC_DUPLEX_UNKNOWN
+from ._common import path_exists_strict
 from ._common import supports_ipv6
 from ._common import usage_percent
 from ._compat import b
+from ._compat import basestring
 from ._compat import long
 from ._compat import PY3
 
@@ -51,6 +56,7 @@ __extra__all__ = [
 
 # --- constants
 
+HAS_SMAPS = os.path.exists('/proc/%s/smaps' % os.getpid())
 HAS_PRLIMIT = hasattr(cext, "linux_prlimit")
 
 # RLIMIT_* constants, not guaranteed to be present on all kernels
@@ -63,13 +69,19 @@ if HAS_PRLIMIT:
 CLOCK_TICKS = os.sysconf("SC_CLK_TCK")
 PAGESIZE = os.sysconf("SC_PAGE_SIZE")
 BOOT_TIME = None  # set later
+# Used when reading "big" files, namely /proc/{pid}/smaps and /proc/net/*.
+# On Python 2, using a buffer with open() for such files may result in a
+# speedup, see: https://github.com/giampaolo/psutil/issues/708
+BIGGER_FILE_BUFFERING = -1 if PY3 else 8192
+LITTLE_ENDIAN = sys.byteorder == 'little'
 if PY3:
     FS_ENCODING = sys.getfilesystemencoding()
+    ENCODING_ERRORS_HANDLER = 'surrogateescape'
 if enum is None:
     AF_LINK = socket.AF_PACKET
 else:
     AddressFamily = enum.IntEnum('AddressFamily',
-                                 {'AF_LINK': socket.AF_PACKET})
+                                 {'AF_LINK': int(socket.AF_PACKET)})
     AF_LINK = AddressFamily.AF_LINK
 
 # ioprio_* constants http://linux.die.net/man/2/ioprio_get
@@ -125,20 +137,75 @@ TimeoutExpired = None
 
 # --- utils
 
-def open_text(fname):
-    """On Python 3 opens a file in text mode by using fs encoding.
+def open_binary(fname, **kwargs):
+    return open(fname, "rb", **kwargs)
+
+
+def open_text(fname, **kwargs):
+    """On Python 3 opens a file in text mode by using fs encoding and
+    a proper en/decoding errors handler.
     On Python 2 this is just an alias for open(name, 'rt').
     """
-    kw = dict(encoding=FS_ENCODING) if PY3 else dict()
-    return open(fname, "rt", **kw)
+    if PY3:
+        # See:
+        # https://github.com/giampaolo/psutil/issues/675
+        # https://github.com/giampaolo/psutil/pull/733
+        kwargs.setdefault('encoding', FS_ENCODING)
+        kwargs.setdefault('errors', ENCODING_ERRORS_HANDLER)
+    return open(fname, "rt", **kwargs)
 
 
 def get_procfs_path():
     return sys.modules['psutil'].PROCFS_PATH
 
 
+def readlink(path):
+    """Wrapper around os.readlink()."""
+    assert isinstance(path, basestring), path
+    path = os.readlink(path)
+    # readlink() might return paths containing null bytes ('\x00')
+    # resulting in "TypeError: must be encoded string without NULL
+    # bytes, not str" errors when the string is passed to other
+    # fs-related functions (os.*, open(), ...).
+    # Apparently everything after '\x00' is garbage (we can have
+    # ' (deleted)', 'new' and possibly others), see:
+    # https://github.com/giampaolo/psutil/issues/717
+    path = path.split('\x00')[0]
+    # Certain paths have ' (deleted)' appended. Usually this is
+    # bogus as the file actually exists. Even if it doesn't we
+    # don't care.
+    if path.endswith(' (deleted)') and not path_exists_strict(path):
+        path = path[:-10]
+    return path
+
+
+def file_flags_to_mode(flags):
+    modes_map = {os.O_RDONLY: 'r', os.O_WRONLY: 'w', os.O_RDWR: 'w+'}
+    mode = modes_map[flags & (os.O_RDONLY | os.O_WRONLY | os.O_RDWR)]
+    if flags & os.O_APPEND:
+        mode = mode.replace('w', 'a', 1)
+    mode = mode.replace('w+', 'r+')
+    # possible values: r, w, a, r+, a+
+    return mode
+
+
+def get_sector_size():
+    try:
+        with open(b"/sys/block/sda/queue/hw_sector_size") as f:
+            return int(f.read())
+    except (IOError, ValueError):
+        # man iostat states that sectors are equivalent with blocks and
+        # have a size of 512 bytes since 2.4 kernels. This value is
+        # needed to calculate the amount of disk I/O in bytes.
+        return 512
+
+
+SECTOR_SIZE = get_sector_size()
+
+
 # --- named tuples
 
+@memoize
 def set_scputimes_ntuple(procfs_path):
     """Return a namedtuple of variable fields depending on the
     CPU times available on this Linux kernel version which may be:
@@ -146,7 +213,7 @@ def set_scputimes_ntuple(procfs_path):
      [guest_nice]]])
     """
     global scputimes
-    with open('%s/stat' % procfs_path, 'rb') as f:
+    with open_binary('%s/stat' % procfs_path) as f:
         values = f.readline().split()[1:]
     fields = ['user', 'nice', 'system', 'idle', 'iowait', 'irq', 'softirq']
     vlen = len(values)
@@ -163,14 +230,26 @@ def set_scputimes_ntuple(procfs_path):
     return scputimes
 
 
-scputimes = set_scputimes_ntuple('/proc')
+try:
+    scputimes = set_scputimes_ntuple("/proc")
+except Exception:
+    # Don't want to crash at import time.
+    traceback.print_exc()
+    scputimes = namedtuple('scputimes', 'user system idle')(0.0, 0.0, 0.0)
+
 
 svmem = namedtuple(
     'svmem', ['total', 'available', 'percent', 'used', 'free',
               'active', 'inactive', 'buffers', 'cached'])
-
-pextmem = namedtuple('pextmem', 'rss vms shared text lib data dirty')
-
+sdiskio = namedtuple('sdiskio', ['read_count', 'write_count',
+                                 'read_bytes', 'write_bytes',
+                                 'read_time', 'write_time',
+                                 'read_merged_count', 'write_merged_count',
+                                 'busy_time'])
+popenfile = namedtuple('popenfile',
+                       ['path', 'fd', 'position', 'mode', 'flags'])
+pmem = namedtuple('pmem', 'rss vms shared text lib data dirty')
+pfullmem = namedtuple('pfullmem', pmem._fields + ('uss', 'pss', 'swap'))
 pmmap_grouped = namedtuple(
     'pmmap_grouped', ['path', 'rss', 'size', 'pss', 'shared_clean',
                       'shared_dirty', 'private_clean', 'private_dirty',
@@ -183,9 +262,16 @@ pmmap_ext = namedtuple(
 # --- system memory
 
 def virtual_memory():
-    total, free, buffers, shared, _, _ = cext.linux_sysinfo()
+    total, free, buffers, shared, _, _, unit_multiplier = cext.linux_sysinfo()
+    total *= unit_multiplier
+    free *= unit_multiplier
+    buffers *= unit_multiplier
+    # XXX: tis is currently not used (neither returned) because it's
+    # always 0. It would be nice to have though ('free' provides it).
+    # shared *= unit_multiplier
+
     cached = active = inactive = None
-    with open('%s/meminfo' % get_procfs_path(), 'rb') as f:
+    with open_binary('%s/meminfo' % get_procfs_path()) as f:
         for line in f:
             if line.startswith(b"Cached:"):
                 cached = int(line.split()[1]) * 1024
@@ -204,6 +290,7 @@ def virtual_memory():
                   "be determined and were set to 0"
             warnings.warn(msg, RuntimeWarning)
             cached = active = inactive = 0
+
     avail = free + buffers + cached
     used = total - free
     percent = usage_percent((total - avail), total, _round=1)
@@ -212,27 +299,40 @@ def virtual_memory():
 
 
 def swap_memory():
-    _, _, _, _, total, free = cext.linux_sysinfo()
+    _, _, _, _, total, free, unit_multiplier = cext.linux_sysinfo()
+    total *= unit_multiplier
+    free *= unit_multiplier
     used = total - free
     percent = usage_percent(used, total, _round=1)
     # get pgin/pgouts
-    with open("%s/vmstat" % get_procfs_path(), "rb") as f:
-        sin = sout = None
-        for line in f:
-            # values are expressed in 4 kilo bytes, we want bytes instead
-            if line.startswith(b'pswpin'):
-                sin = int(line.split(b' ')[1]) * 4 * 1024
-            elif line.startswith(b'pswpout'):
-                sout = int(line.split(b' ')[1]) * 4 * 1024
-            if sin is not None and sout is not None:
-                break
-        else:
-            # we might get here when dealing with exotic Linux flavors, see:
-            # https://github.com/giampaolo/psutil/issues/313
-            msg = "'sin' and 'sout' swap memory stats couldn't " \
-                  "be determined and were set to 0"
-            warnings.warn(msg, RuntimeWarning)
-            sin = sout = 0
+    try:
+        f = open_binary("%s/vmstat" % get_procfs_path())
+    except IOError as err:
+        # see https://github.com/giampaolo/psutil/issues/722
+        msg = "'sin' and 'sout' swap memory stats couldn't " \
+              "be determined and were set to 0 (%s)" % str(err)
+        warnings.warn(msg, RuntimeWarning)
+        sin = sout = 0
+    else:
+        with f:
+            sin = sout = None
+            for line in f:
+                # values are expressed in 4 kilo bytes, we want
+                # bytes instead
+                if line.startswith(b'pswpin'):
+                    sin = int(line.split(b' ')[1]) * 4 * 1024
+                elif line.startswith(b'pswpout'):
+                    sout = int(line.split(b' ')[1]) * 4 * 1024
+                if sin is not None and sout is not None:
+                    break
+            else:
+                # we might get here when dealing with exotic Linux
+                # flavors, see:
+                # https://github.com/giampaolo/psutil/issues/313
+                msg = "'sin' and 'sout' swap memory stats couldn't " \
+                      "be determined and were set to 0"
+                warnings.warn(msg, RuntimeWarning)
+                sin = sout = 0
     return _common.sswap(total, used, free, percent, sin, sout)
 
 
@@ -247,7 +347,7 @@ def cpu_times():
     """
     procfs_path = get_procfs_path()
     set_scputimes_ntuple(procfs_path)
-    with open('%s/stat' % procfs_path, 'rb') as f:
+    with open_binary('%s/stat' % procfs_path) as f:
         values = f.readline().split()
     fields = values[1:len(scputimes._fields) + 1]
     fields = [float(x) / CLOCK_TICKS for x in fields]
@@ -261,7 +361,7 @@ def per_cpu_times():
     procfs_path = get_procfs_path()
     set_scputimes_ntuple(procfs_path)
     cpus = []
-    with open('%s/stat' % procfs_path, 'rb') as f:
+    with open_binary('%s/stat' % procfs_path) as f:
         # get rid of the first line which refers to system wide CPU stats
         f.readline()
         for line in f:
@@ -281,7 +381,7 @@ def cpu_count_logical():
     except ValueError:
         # as a second fallback we try to parse /proc/cpuinfo
         num = 0
-        with open('%s/cpuinfo' % get_procfs_path(), 'rb') as f:
+        with open_binary('%s/cpuinfo' % get_procfs_path()) as f:
             for line in f:
                 if line.lower().startswith(b'processor'):
                     num += 1
@@ -307,7 +407,7 @@ def cpu_count_physical():
     """Return the number of physical cores in the system."""
     mapping = {}
     current_info = {}
-    with open('%s/cpuinfo' % get_procfs_path(), 'rb') as f:
+    with open_binary('%s/cpuinfo' % get_procfs_path()) as f:
         for line in f:
             line = line.strip().lower()
             if not line:
@@ -328,6 +428,26 @@ def cpu_count_physical():
     return sum(mapping.values()) or None
 
 
+def cpu_stats():
+    with open_binary('%s/stat' % get_procfs_path()) as f:
+        ctx_switches = None
+        interrupts = None
+        soft_interrupts = None
+        for line in f:
+            if line.startswith(b'ctxt'):
+                ctx_switches = int(line.split()[1])
+            elif line.startswith(b'intr'):
+                interrupts = int(line.split()[1])
+            elif line.startswith(b'softirq'):
+                soft_interrupts = int(line.split()[1])
+            if ctx_switches is not None and soft_interrupts is not None \
+                    and interrupts is not None:
+                break
+    syscalls = 0
+    return _common.scpustats(
+        ctx_switches, interrupts, soft_interrupts, syscalls)
+
+
 # --- other system functions
 
 def users():
@@ -351,7 +471,7 @@ def users():
 def boot_time():
     """Return the system boot time expressed in seconds since the epoch."""
     global BOOT_TIME
-    with open('%s/stat' % get_procfs_path(), 'rb') as f:
+    with open_binary('%s/stat' % get_procfs_path()) as f:
         for line in f:
             if line.startswith(b'btime'):
                 ret = float(line.strip().split()[1])
@@ -416,8 +536,7 @@ class Connections:
         inodes = defaultdict(list)
         for fd in os.listdir("%s/%s/fd" % (self._procfs_path, pid)):
             try:
-                inode = os.readlink("%s/%s/fd/%s" % (
-                    self._procfs_path, pid, fd))
+                inode = readlink("%s/%s/fd/%s" % (self._procfs_path, pid, fd))
             except OSError as err:
                 # ENOENT == file which is gone in the meantime;
                 # os.stat('/proc/%s' % self.pid) will be done later
@@ -480,7 +599,7 @@ class Connections:
             ip = ip.encode('ascii')
         if family == socket.AF_INET:
             # see: https://github.com/giampaolo/psutil/issues/201
-            if sys.byteorder == 'little':
+            if LITTLE_ENDIAN:
                 ip = socket.inet_ntop(family, base64.b16decode(ip)[::-1])
             else:
                 ip = socket.inet_ntop(family, base64.b16decode(ip))
@@ -492,7 +611,7 @@ class Connections:
             ip = base64.b16decode(ip)
             try:
                 # see: https://github.com/giampaolo/psutil/issues/201
-                if sys.byteorder == 'little':
+                if LITTLE_ENDIAN:
                     ip = socket.inet_ntop(
                         socket.AF_INET6,
                         struct.pack('>4I', *struct.unpack('<4I', ip)))
@@ -513,16 +632,16 @@ class Connections:
         if file.endswith('6') and not os.path.exists(file):
             # IPv6 not supported
             return
-        with open_text(file) as f:
+        with open_text(file, buffering=BIGGER_FILE_BUFFERING) as f:
             f.readline()  # skip the first line
-            for line in f:
+            for lineno, line in enumerate(f, 1):
                 try:
                     _, laddr, raddr, status, _, _, _, _, _, inode = \
                         line.split()[:10]
                 except ValueError:
                     raise RuntimeError(
-                        "error while parsing %s; malformed line %r" % (
-                            file, line))
+                        "error while parsing %s; malformed line %s %r" % (
+                            file, lineno, line))
                 if inode in inodes:
                     # # We assume inet sockets are unique, so we error
                     # # out if there are multiple references to the
@@ -549,15 +668,16 @@ class Connections:
 
     def process_unix(self, file, family, inodes, filter_pid=None):
         """Parse /proc/net/unix files."""
-        # see: https://github.com/giampaolo/psutil/issues/675
-        kw = dict(encoding=FS_ENCODING, errors='replace') if PY3 else dict()
-        with open(file, 'rt', **kw) as f:
+        with open_text(file, buffering=BIGGER_FILE_BUFFERING) as f:
             f.readline()  # skip the first line
             for line in f:
                 tokens = line.split()
                 try:
                     _, _, _, _, type_, _, inode = tokens[0:7]
                 except ValueError:
+                    if ' ' not in line:
+                        # see: https://github.com/giampaolo/psutil/issues/766
+                        continue
                     raise RuntimeError(
                         "error while parsing %s; malformed line %r" % (
                             file, line))
@@ -669,50 +789,70 @@ def disk_io_counters():
     """Return disk I/O statistics for every disk installed on the
     system as a dict of raw tuples.
     """
-    # man iostat states that sectors are equivalent with blocks and
-    # have a size of 512 bytes since 2.4 kernels. This value is
-    # needed to calculate the amount of disk I/O in bytes.
-    SECTOR_SIZE = 512
-
     # determine partitions we want to look for
-    partitions = []
-    with open_text("%s/partitions" % get_procfs_path()) as f:
-        lines = f.readlines()[2:]
-    for line in reversed(lines):
-        _, _, _, name = line.split()
-        if name[-1].isdigit():
-            # we're dealing with a partition (e.g. 'sda1'); 'sda' will
-            # also be around but we want to omit it
-            partitions.append(name)
-        else:
-            if not partitions or not partitions[-1].startswith(name):
-                # we're dealing with a disk entity for which no
-                # partitions have been defined (e.g. 'sda' but
-                # 'sda1' was not around), see:
-                # https://github.com/giampaolo/psutil/issues/338
+    def get_partitions():
+        partitions = []
+        with open_text("%s/partitions" % get_procfs_path()) as f:
+            lines = f.readlines()[2:]
+        for line in reversed(lines):
+            _, _, _, name = line.split()
+            if name[-1].isdigit():
+                # we're dealing with a partition (e.g. 'sda1'); 'sda' will
+                # also be around but we want to omit it
                 partitions.append(name)
-    #
+            else:
+                if not partitions or not partitions[-1].startswith(name):
+                    # we're dealing with a disk entity for which no
+                    # partitions have been defined (e.g. 'sda' but
+                    # 'sda1' was not around), see:
+                    # https://github.com/giampaolo/psutil/issues/338
+                    partitions.append(name)
+        return partitions
+
     retdict = {}
+    partitions = get_partitions()
     with open_text("%s/diskstats" % get_procfs_path()) as f:
         lines = f.readlines()
     for line in lines:
-        # http://www.mjmwired.net/kernel/Documentation/iostats.txt
+        # OK, this is a bit confusing. The format of /proc/diskstats can
+        # have 3 variations.
+        # On Linux 2.4 each line has always 15 fields, e.g.:
+        # "3     0   8 hda 8 8 8 8 8 8 8 8 8 8 8"
+        # On Linux 2.6+ each line *usually* has 14 fields, and the disk
+        # name is in another position, like this:
+        # "3    0   hda 8 8 8 8 8 8 8 8 8 8 8"
+        # ...unless (Linux 2.6) the line refers to a partition instead
+        # of a disk, in which case the line has less fields (7):
+        # "3    1   hda1 8 8 8 8"
+        # See:
+        # https://www.kernel.org/doc/Documentation/iostats.txt
+        # https://www.kernel.org/doc/Documentation/ABI/testing/procfs-diskstats
         fields = line.split()
-        if len(fields) > 7:
-            _, _, name, reads, _, rbytes, rtime, writes, _, wbytes, wtime = \
-                fields[:11]
+        fields_len = len(fields)
+        if fields_len == 15:
+            # Linux 2.4
+            name = fields[3]
+            reads = int(fields[2])
+            (reads_merged, rbytes, rtime, writes, writes_merged,
+                wbytes, wtime, _, busy_time, _) = map(int, fields[4:14])
+        elif fields_len == 14:
+            # Linux 2.6+, line referring to a disk
+            name = fields[2]
+            (reads, reads_merged, rbytes, rtime, writes, writes_merged,
+                wbytes, wtime, _, busy_time, _) = map(int, fields[3:14])
+        elif fields_len == 7:
+            # Linux 2.6+, line referring to a partition
+            name = fields[2]
+            reads, rbytes, writes, wbytes = map(int, fields[3:])
+            rtime = wtime = reads_merged = writes_merged = busy_time = 0
         else:
-            # from kernel 2.6.0 to 2.6.25
-            _, _, name, reads, rbytes, writes, wbytes = fields
-            rtime, wtime = 0, 0
+            raise ValueError("not sure how to interpret line %r" % line)
+
         if name in partitions:
-            rbytes = int(rbytes) * SECTOR_SIZE
-            wbytes = int(wbytes) * SECTOR_SIZE
-            reads = int(reads)
-            writes = int(writes)
-            rtime = int(rtime)
-            wtime = int(wtime)
-            retdict[name] = (reads, writes, rbytes, wbytes, rtime, wtime)
+            rbytes = rbytes * SECTOR_SIZE
+            wbytes = wbytes * SECTOR_SIZE
+            retdict[name] = (reads, writes, rbytes, wbytes, rtime, wtime,
+                             reads_merged, writes_merged, busy_time)
     return retdict
 
 
@@ -758,9 +898,6 @@ def wrap_exceptions(fun):
         try:
             return fun(self, *args, **kwargs)
         except EnvironmentError as err:
-            # support for private module import
-            if NoSuchProcess is None or AccessDenied is None:
-                raise
             # ENOENT (no such file or directory) gets raised on open().
             # ESRCH (no such process) can get raised on read() if
             # process is gone in meantime.
@@ -772,20 +909,6 @@ def wrap_exceptions(fun):
     return wrapper
 
 
-def wrap_exceptions_w_zombie(fun):
-    """Same as above but also handles zombies."""
-    @functools.wraps(fun)
-    def wrapper(self, *args, **kwargs):
-        try:
-            return wrap_exceptions(fun)(self)
-        except NoSuchProcess:
-            if not pid_exists(self.pid):
-                raise
-            else:
-                raise ZombieProcess(self.pid, self._name, self._ppid)
-    return wrapper
-
-
 class Process(object):
     """Linux process implementation."""
 
@@ -806,7 +929,7 @@ class Process(object):
 
     def exe(self):
         try:
-            exe = os.readlink("%s/%s/exe" % (self._procfs_path, self.pid))
+            return readlink("%s/%s/exe" % (self._procfs_path, self.pid))
         except OSError as err:
             if err.errno in (errno.ENOENT, errno.ESRCH):
                 # no such file error; might be raised also if the
@@ -823,28 +946,27 @@ class Process(object):
                 raise AccessDenied(self.pid, self._name)
             raise
 
-        # readlink() might return paths containing null bytes ('\x00').
-        # Certain names have ' (deleted)' appended. Usually this is
-        # bogus as the file actually exists. Either way that's not
-        # important as we don't want to discriminate executables which
-        # have been deleted.
-        exe = exe.split('\x00')[0]
-        if exe.endswith(' (deleted)') and not os.path.exists(exe):
-            exe = exe[:-10]
-        return exe
-
     @wrap_exceptions
     def cmdline(self):
         with open_text("%s/%s/cmdline" % (self._procfs_path, self.pid)) as f:
             data = f.read()
+        if not data:
+            # may happen in case of zombie process
+            return []
         if data.endswith('\x00'):
             data = data[:-1]
         return [x for x in data.split('\x00')]
 
+    @wrap_exceptions
+    def environ(self):
+        with open_text("%s/%s/environ" % (self._procfs_path, self.pid)) as f:
+            data = f.read()
+        return parse_environ_block(data)
+
     @wrap_exceptions
     def terminal(self):
         tmap = _psposix._get_terminal_map()
-        with open("%s/%s/stat" % (self._procfs_path, self.pid), 'rb') as f:
+        with open_binary("%s/%s/stat" % (self._procfs_path, self.pid)) as f:
             tty_nr = int(f.read().split(b' ')[6])
         try:
             return tmap[tty_nr]
@@ -855,7 +977,7 @@ class Process(object):
         @wrap_exceptions
         def io_counters(self):
             fname = "%s/%s/io" % (self._procfs_path, self.pid)
-            with open(fname, 'rb') as f:
+            with open_binary(fname) as f:
                 rcount = wcount = rbytes = wbytes = None
                 for line in f:
                     if rcount is None and line.startswith(b"syscr"):
@@ -878,28 +1000,27 @@ class Process(object):
 
     @wrap_exceptions
     def cpu_times(self):
-        with open("%s/%s/stat" % (self._procfs_path, self.pid), 'rb') as f:
+        with open_binary("%s/%s/stat" % (self._procfs_path, self.pid)) as f:
             st = f.read().strip()
         # ignore the first two values ("pid (exe)")
         st = st[st.find(b')') + 2:]
         values = st.split(b' ')
         utime = float(values[11]) / CLOCK_TICKS
         stime = float(values[12]) / CLOCK_TICKS
-        return _common.pcputimes(utime, stime)
+        children_utime = float(values[13]) / CLOCK_TICKS
+        children_stime = float(values[14]) / CLOCK_TICKS
+        return _common.pcputimes(utime, stime, children_utime, children_stime)
 
     @wrap_exceptions
     def wait(self, timeout=None):
         try:
             return _psposix.wait_pid(self.pid, timeout)
         except _psposix.TimeoutExpired:
-            # support for private module import
-            if TimeoutExpired is None:
-                raise
             raise TimeoutExpired(timeout, self.pid, self._name)
 
     @wrap_exceptions
     def create_time(self):
-        with open("%s/%s/stat" % (self._procfs_path, self.pid), 'rb') as f:
+        with open_binary("%s/%s/stat" % (self._procfs_path, self.pid)) as f:
             st = f.read().strip()
         # ignore the first two values ("pid (exe)")
         st = st[st.rfind(b')') + 2:]
@@ -914,13 +1035,6 @@ class Process(object):
 
     @wrap_exceptions
     def memory_info(self):
-        with open("%s/%s/statm" % (self._procfs_path, self.pid), 'rb') as f:
-            vms, rss = f.readline().split()[:2]
-            return _common.pmem(int(rss) * PAGESIZE,
-                                int(vms) * PAGESIZE)
-
-    @wrap_exceptions
-    def memory_info_ex(self):
         #  ============================================================
         # | FIELD  | DESCRIPTION                         | AKA  | TOP  |
         #  ============================================================
@@ -932,12 +1046,45 @@ class Process(object):
         # | data   | data + stack                        | drs  | DATA |
         # | dirty  | dirty pages (unused in Linux 2.6)   | dt   |      |
         #  ============================================================
-        with open("%s/%s/statm" % (self._procfs_path, self.pid), "rb") as f:
+        with open_binary("%s/%s/statm" % (self._procfs_path, self.pid)) as f:
             vms, rss, shared, text, lib, data, dirty = \
                 [int(x) * PAGESIZE for x in f.readline().split()[:7]]
-        return pextmem(rss, vms, shared, text, lib, data, dirty)
+            return pmem(rss, vms, shared, text, lib, data, dirty)
 
-    if os.path.exists('/proc/%s/smaps' % os.getpid()):
+    # /proc/pid/smaps does not exist on kernels < 2.6.14 or if
+    # CONFIG_MMU kernel configuration option is not enabled.
+    if HAS_SMAPS:
+
+        @wrap_exceptions
+        def memory_full_info(
+                self,
+                _private_re=re.compile(b"Private.*:\s+(\d+)"),
+                _pss_re=re.compile(b"Pss.*:\s+(\d+)"),
+                _swap_re=re.compile(b"Swap.*:\s+(\d+)")):
+            basic_mem = self.memory_info()
+            # Note: using 3 regexes is faster than reading the file
+            # line by line.
+            # XXX: on Python 3 the 2 regexes are 30% slower than on
+            # Python 2 though. Figure out why.
+            with open_binary("%s/%s/smaps" % (self._procfs_path, self.pid),
+                             buffering=BIGGER_FILE_BUFFERING) as f:
+                smaps_data = f.read()
+            # You might be tempted to calculate USS by subtracting
+            # the "shared" value from the "resident" value in
+            # /proc/<pid>/statm. But at least on Linux, statm's "shared"
+            # value actually counts pages backed by files, which has
+            # little to do with whether the pages are actually shared.
+            # /proc/self/smaps on the other hand appears to give us the
+            # correct information.
+            uss = sum(map(int, _private_re.findall(smaps_data))) * 1024
+            pss = sum(map(int, _pss_re.findall(smaps_data))) * 1024
+            swap = sum(map(int, _swap_re.findall(smaps_data))) * 1024
+            return pfullmem(*basic_mem + (uss, pss, swap))
+
+    else:
+        memory_full_info = memory_info
+
+    if HAS_SMAPS:
 
         @wrap_exceptions
         def memory_maps(self):
@@ -945,7 +1092,8 @@ class Process(object):
             Fields are explained in 'man proc'; here is an updated (Apr 2012)
             version: http://goo.gl/fmebo
             """
-            with open_text("%s/%s/smaps" % (self._procfs_path, self.pid)) as f:
+            with open_text("%s/%s/smaps" % (self._procfs_path, self.pid),
+                           buffering=BIGGER_FILE_BUFFERING) as f:
                 first_line = f.readline()
                 current_block = [first_line]
 
@@ -982,6 +1130,9 @@ class Process(object):
                             path = '[anon]'
                         else:
                             path = path.strip()
+                            if (path.endswith(' (deleted)') and not
+                                    path_exists_strict(path)):
+                                path = path[:-10]
                         ls.append((
                             addr, perms, path,
                             data['Rss:'],
@@ -997,25 +1148,14 @@ class Process(object):
                         ))
             return ls
 
-    else:
-        def memory_maps(self):
-            msg = "couldn't find /proc/%s/smaps; kernel < 2.6.14 or "  \
-                  "CONFIG_MMU kernel configuration option is not enabled" \
-                  % self.pid
-            raise NotImplementedError(msg)
-
-    @wrap_exceptions_w_zombie
+    @wrap_exceptions
     def cwd(self):
-        # readlink() might return paths containing null bytes causing
-        # problems when used with other fs-related functions (os.*,
-        # open(), ...)
-        path = os.readlink("%s/%s/cwd" % (self._procfs_path, self.pid))
-        return path.replace('\x00', '')
+        return readlink("%s/%s/cwd" % (self._procfs_path, self.pid))
 
     @wrap_exceptions
     def num_ctx_switches(self):
         vol = unvol = None
-        with open("%s/%s/status" % (self._procfs_path, self.pid), "rb") as f:
+        with open_binary("%s/%s/status" % (self._procfs_path, self.pid)) as f:
             for line in f:
                 if line.startswith(b"voluntary_ctxt_switches"):
                     vol = int(line.split()[1])
@@ -1030,7 +1170,7 @@ class Process(object):
 
     @wrap_exceptions
     def num_threads(self):
-        with open("%s/%s/status" % (self._procfs_path, self.pid), "rb") as f:
+        with open_binary("%s/%s/status" % (self._procfs_path, self.pid)) as f:
             for line in f:
                 if line.startswith(b"Threads:"):
                     return int(line.split()[1])
@@ -1046,7 +1186,7 @@ class Process(object):
             fname = "%s/%s/task/%s/stat" % (
                 self._procfs_path, self.pid, thread_id)
             try:
-                with open(fname, 'rb') as f:
+                with open_binary(fname) as f:
                     st = f.read().strip()
             except IOError as err:
                 if err.errno == errno.ENOENT:
@@ -1170,7 +1310,7 @@ class Process(object):
 
     @wrap_exceptions
     def status(self):
-        with open("%s/%s/status" % (self._procfs_path, self.pid), 'rb') as f:
+        with open_binary("%s/%s/status" % (self._procfs_path, self.pid)) as f:
             for line in f:
                 if line.startswith(b"State:"):
                     letter = line.split()[1]
@@ -1188,7 +1328,7 @@ class Process(object):
         for fd in files:
             file = "%s/%s/fd/%s" % (self._procfs_path, self.pid, fd)
             try:
-                file = os.readlink(file)
+                path = readlink(file)
             except OSError as err:
                 # ENOENT == file which is gone in the meantime
                 if err.errno in (errno.ENOENT, errno.ESRCH):
@@ -1200,12 +1340,19 @@ class Process(object):
                 else:
                     raise
             else:
-                # If file is not an absolute path there's no way
-                # to tell whether it's a regular file or not,
-                # so we skip it. A regular file is always supposed
-                # to be absolutized though.
-                if file.startswith('/') and isfile_strict(file):
-                    ntuple = _common.popenfile(file, int(fd))
+                # If path is not an absolute there's no way to tell
+                # whether it's a regular file or not, so we skip it.
+                # A regular file is always supposed to be have an
+                # absolute path though.
+                if path.startswith('/') and isfile_strict(path):
+                    # Get file position and flags.
+                    file = "%s/%s/fdinfo/%s" % (
+                        self._procfs_path, self.pid, fd)
+                    with open_binary(file) as f:
+                        pos = int(f.readline().split()[1])
+                        flags = int(f.readline().split()[1], 8)
+                    mode = file_flags_to_mode(flags)
+                    ntuple = popenfile(path, int(fd), int(pos), mode, flags)
                     retlist.append(ntuple)
         if hit_enoent:
             # raise NSP if the process disappeared on us
@@ -1226,7 +1373,7 @@ class Process(object):
     @wrap_exceptions
     def ppid(self):
         fpath = "%s/%s/status" % (self._procfs_path, self.pid)
-        with open(fpath, 'rb') as f:
+        with open_binary(fpath) as f:
             for line in f:
                 if line.startswith(b"PPid:"):
                     # PPid: nnnn
@@ -1236,7 +1383,7 @@ class Process(object):
     @wrap_exceptions
     def uids(self):
         fpath = "%s/%s/status" % (self._procfs_path, self.pid)
-        with open(fpath, 'rb') as f:
+        with open_binary(fpath) as f:
             for line in f:
                 if line.startswith(b'Uid:'):
                     _, real, effective, saved, fs = line.split()
@@ -1246,7 +1393,7 @@ class Process(object):
     @wrap_exceptions
     def gids(self):
         fpath = "%s/%s/status" % (self._procfs_path, self.pid)
-        with open(fpath, 'rb') as f:
+        with open_binary(fpath) as f:
             for line in f:
                 if line.startswith(b'Gid:'):
                     _, real, effective, saved, fs = line.split()
